PROGRAM CPULoader16_32;
(*
================================================================================================================
Author:
   This CPULoader / MMCLoader for AVRco Pascal has originally been written by Robert Sander for FAT16.

================================================================================================================
Changes:

   The following MAJOR changes have been made by Michael Bremicker Oct.2012:

       - The MMCLoader Unit has been renamed to MMCLoader16_32
       - It detects and supports MMC-, SD- and SDHC-Cards
       - It detects and supports FAT16 and FAT32 file systems
       - The Version- and Checksum-Procedure have been REMOVED due to space requirements - check this in the main application before calling the CPU-Loader!
       - The complete Unit is documented now. It has been heavily re-structured and re-written to save space and to simplify it.

   The following changes have been made by Michael Bremicker Sept.2012:

       - This CPULoader (Bootloader) can be updated by the main application now.
       - Checksum and Loader Version/Loader Flag Record implemented into the CPULoader in the header of the Unit MMCLoader.pas
       - Compiled Filesize further compressed, still below 4 K (Using the Merlin Optimizer)
       - BOOTSZ Fuses changed: some of the ATMega Bootloader area must be reserved for the Main Application's BootLoader Updater. Only the ATMega's Bootloader area allows the execution of the SPM command,
         which is mandatory for flash writing!! So, we use a 8 K Bootloader area, but the first 4 K are available for the Main Application, the second 4 K for this Bootloader.

   The following changes have been made by Michael Bremicker Oct.2009:
   
       - Reduction to the neccessary minimum, to fit into 4 kB Boot flash when compiling with the merlin optimizer
       - Automatic checksum function for the update file
       - Possibility to jump back from the flashed runnning application to do another flash update (!)
       - Fast Startup on PowerUp or Hardware-Reset

   Please note the {!} that are in the source - here you may have to do changes according the system you use!
================================================================================================================

   Used Terms:
   
        1. MCU:          Micro Controller Unit
        2. MMC:          Multi Media Card, a memory-card, here up to 32 GB space, FAT16 (max. 2 GByte) or FAT32 formatted
        3. CPULoader:    This software including the MMCLoader16_32 unit, residing in the bootloader area of an ATMega MCU.
        4. Application:  Another software, residing in the normal program flash area of an ATMega MCU,
                         starting normally at address $0000.


   What this software can do:
   
        1. This program is a CPULoader, which can read a binary file (e.g. \CPU.BIN directly written by AVRco's programmer to the MMC's root)
           from a FAT16 / 32 MMC / SD / SDHC Card and flash it to the MCU
           This file must be as least as big as the complete Application memory area (Flash - BootSize).
           AVRco's programmer makes it as big as the complete flash - which is no harm and has been taken care of.
        2. This program can update a running application by leaving the application and invoking the CPULoader again.
           So, the application can provide a new update file, e.g. by serial interface or ethernet, write it to the MMC
           and call the CPULoader then - thus updating the application.
        3. This program executes the flash update, if the Update File exists.


   What hardware is required:

        1. An ATMega MCU with a boot area of at least 4 kB. (You MUST compile it with the MERLIN OPTIMIZER to stay under 4 KB!!)
        2. A FAT16 / 32 formatted MMC / SD / SDHC Card attached the the MCU via the hardware SPI of the MCU.
        

   What must be done to use this CPULoader:
   
       1. Adapt the {!} locations according to your need (Do NOT forget the changes the MMCLoader16_32 unit!)
          When compiled with the Merlin Optimizer the CPULoader needs a little less than 4 kB of space,
          so it can reside in the bootarea of smaller ATMegas too.
       2. Flash the CPULoader into your ATMega to the boot section.
       3. Prepare your main program (Application) for the use with this CPULoader.
          See our example project Appl4Loader.Pas to find the requirements!
       4. Compile your main program (Application) in AVRco, open the AVRco programmer,
          select 'File' --> 'Save Flash Buffer to binary File' and save CPU.BIN to the root of your MMC.
       5. Now insert the MMC / SD / SDHC into your MCU system. It should flash your main program automatically.
       6. You can upload a newer Application file to your MMC / SD / SDHC Card and invoke the flashing from your application.
       7. When you do a Power Up or a hardware reset NO check for updating is done for reason of fast startup.
          Check for update is done only, when no valid application is loaded or when the application calls the CPULoader
       
       
   Speed measurement in my system (ATMega 2561, 14.74 MHz, 256 kB Flash, 2GB MMC):

       1. Reading the File (256 kB) for calculating the checksum and getting the Build Number: 4 seconds
       2. Reading the File again and flashing the required 248 kB application: 11 seconds

*)


{$NOSHADOW}
{ $WG}                     {global Warnings off}
{$DEFINE MMC_SDV1_Support}                                                            // If defined, the Bootloader will support MMC-Cards too. It increases the Size roughly by 200 Bytes
{!}
Device = mega2561, VCC=5;

{!}
{$BootApplication $1F800}                                                       // This CPULoader runs completely in the Boot area!
                                                                                // Set the Address (in WORDS) according the Fuses and the Controller
                                                                                //
                                                                                // If you put the CPULoader to $1F800 (ONLY 4 KB of Boot-Memory) you MUST compile it with the
                                                                                // MERLIN OPTIMIZER in order to keep it under 4 KB!
                                                                                //
Define_Fuses                                                                    // Fuses to be set. Must be in accordance with the hardware AND the application
  Override_Fuses;                                                               // Fuse listed here = Fuse Set = Value 0  !!
  LockBits0      = [];                                                          //
{!}
  FuseBits0      = [SUT1];                                                      // CKSEL3..1 = 111 --> 16 MHz external oscillator
                                                                                // CKSEL0 = 1, SUT1..0 = 01 --> Crystal oscillator, BOD enabled
{!}
  FuseBits1      = [SPIEN, EESAVE, BOOTSZ0,BOOTSZ1];                            // BOOTSZ1\SZ0 0 0 --> Bootloader at WORD ADDRESS $1F000, EESAVE --> EEPROM is preserved during a chip erase
                                                                                // CAVE: We have a Dirty Trick here! We place the BootLoader Code at Address $1F800, but we set both BOOTSZ0/1 indicating 1F000 for the boot area!!
                                                                                // This is to use the range from $1F000 to $1F800 in the Main Application (starting $00000) as BootLoader Flashing area for BootLoader Updates!
                                                                                // BOOTRST not (!!!) set --> Jump to $0000 (Application) and not to the Bootloader when Power Up or Reset
  FuseBits2      = [BODLEVEL0,BODLEVEL2];                                       // Brown Out at 4.3V BL0: 1, BL1: 1, BL2: 0;     2.7V: BL0: 0, BL1: 1, BL2: 0
  ProgMode       = SPI;

  ProgFuses   = false;  // or false  program Fuse Bits (* REV4 *)
  ProgLock    = false;  // or false  program Lock Bits (* REV4 *)
  ProgFlash   = true;   // or false  program Flash (* REV4 *)
  ProgEEprom  = false;  // or false  program Eeprom (* REV4 *)

IMPORT
  FlashWrite;

FROM System IMPORT LONGWORD, FlashCheck_S;

Define
{!}
  FlashChkSum = $3FFFE;                                                         // Byte Address. Here the Compiler writes the CheckSum (Word!) of the Compiled binary data
  ProcClock      = 14745600;       {Hertz}
  StackSize      = $0200, iData;
  FrameSize      = $0200, iData;


USES
  MMCLoader16_32;

IMPLEMENTATION

{$IDATA}


Var
   gPowerUp : Boolean;                                                          // global variable to remember the status of the system, when this routine is entered
Begin
   gPowerUp := ((MCUSR And 1) = 1);                                             // MCU STATUS Register: If the System has a real Power Up, Bit 0 in the MCUSR is set. Remember it!
   EXCL(MCUSR,0);                                                               // After that clear the bit
   If Not gPowerUp And ((MCUSR And 2) = 2) Then                                 // In case we have NO Power Up, check for an external (hardware) Reset: Remember it too.
      gPowerUp := True;                                                         // If gPowerUp is True, we just check for a valid application and jump immediately to it
   EndIf;                                                                       // If gPowerUp is False, we search the MMC card for a valid update file and check its version and checksum
                                                                                // gPowerUp is False, if the application (that wants a flash update) calls the CPULoader
   INCL(MCUCR,1);                                                               // MCU CONTROL Register: set the IVSEL (Interrupt vector table selection) to the boot area!
                                                                                //

   If (BootFlag.Flag = $AA55) And gPowerUp Then                                 // The flag $AA55 indicates a properly flashed MC-Hx Application in the application area.
      EXCL(MCUCR,1);                                                            // If we have the correct Flag AND a fresh Power Up or Hardware Reset:
      Application_Startup;                                                      // Then switch back to Application Vector Table and go directly to the Application
   EndIf;                                                                       // So, we do NOT check the MMC Card for a valid update file. This speeds up the Power Up / Reset startup process

   DisableInts;                                                                 // We do not use any interrupts

   mDelay(300);                                                                 // Very important! Allow some time for the MMC Card to get ready.
   Loop                                                                         // If there is no proper flashed Application or the BootLoader was called from the Application: Check:
     If FATInit Then                                                            // Is a valid FAT 16 / 32 System available?
        UpdateFirmware;                                                         // If yes: go here and flash the file
        EXCL(MCUCR,1);                                                          // MCU CONTROL Register: set the IVSEL (Interrupt vector table selection) to the application area!
        Application_Startup;                                                    // then start up the application. The Application SHOULD ALWAYS run a SYSTEM_RESET once after an update!
                                                                                // The TWI and the MMC are otherwise not proplerly initiated!
                                                                                // You can NOT run a SYSTEM_RESET here!
     EndIf;
  EndLoop;
End CPULoader16_32.

